#!/usr/bin/env python3
# Based on https://zenhax.com/viewtopic.php?f=6&t=3324

import sys
import wave
import argparse
import pathlib

from wgrd_cons_tools.utils import *


def decodeBlock(f, num19, array4, num7):

  array3 = [0] * num7
  array2 = [0] * num7

  num9 = read16sb(f)
  num10 = read16sb(f)
  num11 = read16sb(f)
  num12 = read16sb(f)
  num13 = read16sb(f)
  num14 = read16sb(f)
  num15 = read16sb(f)
  num16 = read16sb(f)
  num17 = read16sb(f)
  num18 = read16sb(f)
  num20 = num19 - 1

  sampleIndex = 0
  while sampleIndex < num7:
    num21 = (array4[num19 // 8] >> 7 - num19 % 8) & 1
    if (num21 > 0):
      num22 = 0
      num23 = num19 - num20 - 1
      num20 = num19
      if (num23 >= 2):
        if (num23 >= 4):
          if (num23 >= 6):
            if (num23 == 24):
              num24 = 0
              for k in range(16):
                num19 += 1
                num24 = num24 * 2 + ((array4[num19 // 8] >> 7 - num19 % 8) & 1)
                num20 += 1
              num23 = num24 + 24

            num25 = num18
            num26 = num17
            num22 = num16 + num26 + (num25 + 1) * ((num23 >> 1) - 2)
            num8 = num25 + num22
            num16 += 6 * ((num16 + 2048) // 2048)
            num17 = num26 + 6 * ((num26 + 1024) // 1024)
            num18 = num25 + 6 * ((num25 + 512) // 512)
          else:
            num27 = num17
            num28 = num18
            num22 = num27 + num16
            num8 = num28 + num22
            num16 += 6 * ((num16 + 2048) // 2048)
            num17 = num27 + 6 * ((num27 + 1024) // 1024)
            num18 = num28 - 2 * ((num28 + 510) // 512)
        else:
          num22 = num16
          num29 = num17
          num8 = num29 + num16
          num16 = num22 + 6 * ((num22 + 2048) // 2048)
          num17 = num29 - 2 * ((num29 + 1022) // 1024)
        num30 = num8
      else:
        num30 = num16
        num16 -= 2 * ((num16 + 2046) // 2048)
      num31 = num30 + num22 >> 1

      # FIXME: Simplify?
      if (num31 > num16):
        if (num31 > num17):
          if (num31 <= num18):
            num18 -= 2 * ((num18 + 510) // 512)
        else:
          num17 -= 2 * ((num17 + 1022) // 1024)
      else:
        num16 -= 2 * ((num16 + 2046) // 2048)

      num32 = num30 - num22
      if (num32 > 64):
        num33 = num32 // 4 + 1
        num19 += 1
        num21 = (array4[num19 // 8] >> 7 - num19 % 8) & 1
        num19 += 1
        num34 = num21 * 2 + ((array4[num19 // 8] >> 7 - num19 % 8) & 1)
        num20 += 2
        num31 = num22 + num33 * num34
      if ((num23 & 1) > 0):
        num31 = ~num31
      array3[sampleIndex] = num31
      sampleIndex += 1
    num19 += 1

  for sampleIndex in range(num7):
    num35 = array3[sampleIndex]
    num27 = num11
    num36 = 2 * (2 * num13 - num15) - 5 * num14
    num37 = 2 * num11 - num12
    num38 = num14
    num14 = num13
    num13 = num35 + (num10 * num36 + 128 >> 8)

    # ???
    if ((num36 | num35) != 0):
      num10 += (((num36 ^ num35) & -536870913) | 0x40000000) >> 29

    sample = num13 + (num9 * num37 + 128 >> 8)

    # Clamp sample
    if (sample > 32767):
      sample = 32767
    if (sample < -32768):
      sample = -32768

    # ???
    if ((num13 | num37) != 0):
      num9 += (((num13 ^ num37) & -536870913) | 0x40000000) >> 29

    num12 = num27
    array2[sampleIndex] = sample
    num11 = sample
    num15 = num38

  return num19, array2

if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("path", type=pathlib.Path, help="path to the ess file")
  parser.add_argument("-o", "--output", type=pathlib.Path, default="test.ess.wav",
                           help="path to the output wav file")
  args = parser.parse_args()

  #print("Loading '%s'" % path)

  data = open(args.path, "rb").read()
  f = BytesIO(data)

  unk0 = read32b(f)
  unk1 = read8(f)
  channelCount = read8(f)
  samplerate = read16b(f)
  frameCount1 = read32b(f)
  unk2 = read32b(f)
  frameCount2 = read32b(f)
  #print("ESS unk0=0x%08X unk1=0x%X channelCount=%d samplerate=%dHz frameCount1=%d unk2=0x%X frameCount2=%d" % (unk0, unk1, channelCount, samplerate, frameCount1, unk2, frameCount2))
  assert(unk0 == 0x01000202)

  # These are mostly 0, I assume it's the number of bits
  assert(unk1 in [0, 1])

  #FIXME: Many values, see sformat.py for more
  #assert(unk2 in [0x0, 0xC0])

  # These are mostly the same, but sometimes frameCount1 is higher?!
  # Number of blocks seems to match frameCount1
  # If this wasn't the case, I'd suspect that frameCount1 is the decoded length (including silence)
  assert(frameCount1 >= frameCount2)

  if channelCount == 1:
    blockMaxFrameCount = 1024 // 1
  elif channelCount == 2:
    blockMaxFrameCount = 1024 // 2
  elif channelCount == 6:
    # All files I've seen are actually 3x stereo tracks
    # The generated files work fine in Audacity, but other tools might have limited support
    blockMaxFrameCount = 1024 // 6
  else:
    assert(False)

  fo = open(args.output, "wb")
  wav = wave.open(fo, 'wb')
  wav.setsampwidth(2)
  wav.setnchannels(channelCount)
  wav.setframerate(samplerate)
  wav.setnframes(frameCount1)

  # Calculate number of blocks
  blockCount = (frameCount1 + (blockMaxFrameCount - 1)) // blockMaxFrameCount
  #print("Number of blocks: %d" % blockCount)

  # Store block offsets
  blockOffsets = [0]
  for blockIndex in range(blockCount):
    blockOffset = read32b(f)
    blockOffsets += [blockOffset]

  dataOffset = f.tell()

  for blockIndex in range(blockCount):
    #print("%d / %d" % (blockIndex, blockCount))
    blockDataSize = blockOffsets[blockIndex + 1] - blockOffsets[blockIndex] - 20 * channelCount

    blockFrameCount = min(frameCount1, blockMaxFrameCount)

    blockOffset = blockOffsets[blockIndex]
    f.seek(dataOffset + blockOffset + 20 * channelCount)
    blockData = f.read(blockDataSize)
    f.seek(dataOffset + blockOffset)

    blockDataOffset = 0
    channelsData = []
    for channelIndex in range(channelCount):
      blockDataOffset, channelData = decodeBlock(f, blockDataOffset, blockData, blockFrameCount)
      channelsData += [channelData]

    # Interleave audio
    frameData = b''
    for frameIndex in range(blockFrameCount):
      for channelIndex in range(channelCount):
        frameData += encode16s(channelsData[channelIndex][frameIndex])

    wav.writeframes(frameData)

    frameCount1 -= blockFrameCount



import ctypes
from io import BytesIO
import zlib
import os
from functools import partial


def debugPrint(s):
    if True:
        return
    print(s)


def remap(value, fromMin, fromMax, toMin, toMax):
    value01 = (value - fromMin) / (fromMax - fromMin)
    return value01 * (toMax - toMin) + toMin

def ctypes_is_int(T):
    return T in [ctypes.c_uint8, ctypes.c_uint16, ctypes.c_uint32, ctypes.c_int8, ctypes.c_int16, ctypes.c_int32]


def ctypes_is_float(T):
    return T in [ctypes.c_float, ctypes.c_double]


def read(f, T):
    data = f.read(ctypes.sizeof(T))
    return T.from_buffer_copy(data, 0)

def readb(f, T):
    data = f.read(ctypes.sizeof(T))
    return T.from_buffer_copy(data[::-1], 0)

def readItems(f, T, c):
    ret = []
    for i in range(c):
        ret.append(read(f, T))
    return ret


def printLoc(f, target):
    c = f.tell()
    print("%d / %d (%+d)" % (c, target, c - target))


def dumpBytes(path, d, output_folder="out/"):
    path = output_folder + path
    abspath = os.path.abspath(path)
    dirname = os.path.dirname(abspath)
    os.makedirs(dirname, exist_ok=True)
    o = open(path, "wb")
    o.write(d)


def dump(f, path, offset, size, output_folder="out/"):
    c = f.tell()
    f.seek(offset)
    d = f.read(size)
    dumpBytes(path, d, output_folder=output_folder)
    f.seek(c)
    return d


def assertLoc(f, target):
    printLoc(f, target)
    c = f.tell()
    assert (c == target)


def readMagic(f):
    return f.read(4)


def writeMagic(f, m):
    assert (len(m) == 4)
    f.write(m)


def readBytes(f, offset, size):
    f.seek(offset)
    d = f.read(size)
    return d

def read16sb(f):
    return readb(f, ctypes.c_int16).value

def read16b(f):
    return readb(f, ctypes.c_uint16).value

def read32b(f):
    return readb(f, ctypes.c_uint32).value

def read32sb(f):
    return readb(f, ctypes.c_int32).value

def read32(f):
    return read(f, ctypes.c_uint32).value


def read32s(f):
    return read(f, ctypes.c_int32).value


def read16(f):
    return read(f, ctypes.c_uint16).value


def read16s(f):
    return read(f, ctypes.c_int16).value


def read8(f):
    return read(f, ctypes.c_uint8).value


def read8s(f):
    return read(f, ctypes.c_int8).value


def readFloat(f):
    return read(f, ctypes.c_float).value


def readDouble(f):
    return read(f, ctypes.c_double).value


def encode8(v):
    return bytes(ctypes.c_uint8(v))


def encode8s(v):
    return bytes(ctypes.c_int8(v))


def encode16(v):
    return bytes(ctypes.c_uint16(v))


def encode16s(v):
    return bytes(ctypes.c_int16(v))


def encode32(v):
    return bytes(ctypes.c_uint32(v))


def encode32s(v):
    return bytes(ctypes.c_int32(v))


def encodeFloat(v):
    return bytes(ctypes.c_float(v))


def encodeDouble(f, v):
    return bytes(ctypes.c_double(v))


def write8(f, v):
    return f.write(encode8(v))


def write8s(f, v):
    return f.write(encode8s(v))


def write16(f, v):
    return f.write(encode16(v))

def write16s(f, v):
    return f.write(encode16s(v))

def write16b(f, v):
    return f.write(encode16(v)[::-1])

def write16sb(f, v):
    return f.write(encode16s(v)[::-1])


def write32(f, v):
    return f.write(encode32(v))

def write32s(f, v):
    return f.write(encode32s(v))

def write32b(f, v):
    return f.write(encode32(v)[::-1])

def write32sb(f, v):
    return f.write(encode32s(v)[::-1])


def writeFloat(f, v):
    return f.write(encodeFloat(v))


def writeDouble(f, v):
    return f.write(encodeDouble(v))


def align(v, n):
    p = (n - (v % n)) % n
    return v + p


def assertPad(b):
    assert(b == b"\x00" * len(b))

def readPad(f, v, n):
    p = align(v, n) - v
    d = f.read(p)
    assert (d == b'\x00' * p)


def writePad(f, v, n):
    p = align(v, n) - v
    d = f.write(b'\x00' * p)


def readString(f):
    s = b''
    while True:
        c = f.read(1)
        if c == b'\x00':
            break
        s += c
    return s


def readPaddedString(f, n):
    s = f.read(n)
    k = s.find(b'\x00')
    if k == -1:
        k = len(s)
    assert (s[k:] == b'\x00' * (n - k))
    return s[0:k]


def asciiSort(l):
    s = []

    def R(a, b):
        l = []
        for x in range(a[0], b[0] + 1):
            l += [bytes([x])]
        return l

    # Allowed symbols and their suspected precedence
    rank = [b'\\'] + [b'-'] + [b'.'] + R(b'0', b'9') + [b'_'] + R(b'a', b'z')

    # print(rank)
    for x in l:
        c = bytes([x])
        # print(c)
        assert (c in rank)
        s += [rank.index(c)]
    return s


def compress_zlib(data, level=9, wbits=15, depth=0):
    print("  " * depth + "compressing zlib data")
    d = data
    io = BytesIO(d)
    data = BytesIO()
    z = zlib.compressobj(level=level, wbits=wbits)
    for block in iter(partial(io.read, 8192), b''):
        data.write(z.compress(block))
    data.write(z.flush(zlib.Z_FULL_FLUSH))
    return bytes(data.getbuffer())


def decompress_zlib(data, wbits=0, depth=0):
    d = data
    fi = BytesIO(d)

    g = b''
    z = zlib.decompressobj(wbits)
    while True:
        buf = z.unconsumed_tail
        if buf == b"":
            buf = fi.read(8192)
            if buf == b"":
                break
        got = z.decompress(buf)
        if got == b"":
            break
        g += got
        # print(g)
    return g


def delta_encode(data, depth=0):
    print("  " * depth + "delta encoding data")
    input = data
    last = 0
    output = BytesIO()
    for block in iter(partial(input.read, 2), b''):
        current = int.from_bytes(block, "little")
        write16s(output, current - last)
        last = current
    return output.getbuffer()


def delta_decode(data, depth=0):
    print("  " * depth + "delta decoding data")
    input = data
    stream = BytesIO(input)
    last = 0
    output = BytesIO()
    for block in iter(partial(stream.read, 2), b''):
        last += int.from_bytes(block, "little")
        write16s(output, last)
    return output.getbuffer()


def writeString(f, s):
    f.write(s + b'\x00')


def readStringAlign(f, n):
    s = readString(f)
    readPad(f, len(s) + 1, n)
    return s


def writeStringAlign(f, n, s):
    writeString(f, s)
    writePad(f, len(s) + 1, n)


# Python `readlines` would keep newlines; naive `split` breaks for zero-length input
def loadLines(path):
    d = open(path, "rb").read()
    if len(d) == 0:
        return []
    return d.split(b'\n')


# Python `writelines` wouldn't include newlines
def dumpLines(path, lines):
    dumpBytes(path, b"\n".join(lines))



#!/usr/bin/env python3

import pdb

import os
import sys
import argparse
import pathlib
import sys

def unpack_file(module, outpath, filepath):
    ret = os.system(f"python3 -m wgrd_cons_parsers.{module} -o '{outpath}' '{filepath}'")
    if ret != 0:
        print(f"failed at {filepath}")
        exit(ret)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("wgrd_path", type=pathlib.Path, help="path to the wgrd game files")
    parser.add_argument("output_path", type=pathlib.Path, help="unpack directory", nargs='?', default="./out/")

    args = parser.parse_args()

    for root, dirs, files in os.walk(args.wgrd_path):
        for file in files:
            if file.endswith(".dat"):
                filepath = os.path.join(root, file)
                outpath = os.path.join(args.output_path, os.path.relpath(filepath, args.wgrd_path))
                #print()
                #print(filepath)

                #os.system(f"python3 -m wgrd_cons_parsers.edat -o '{outpath}' '{filepath}'")

    # we do not need full recursion, because there're no dat files in dats except mpk
    for root, dirs, files in os.walk(args.output_path):
        for file in files:
            if file.endswith(".mpk"):
                filepath = os.path.join(root, file)
                outpath = os.path.dirname(filepath)
                #print(filepath)
                #print(outpath)

                #os.system(f"python3 -m wgrd_cons_parsers.edat -o '{outpath}' '{filepath}'")
    # here all files are already unpacked
    for root, dirs, files in os.walk(args.output_path):
        for file in files:
            filepath = os.path.join(root, file)
            outpath = os.path.dirname(filepath)
            if file.endswith(".ndfbin"):
                pass
            if file.endswith(".dic"):
                pass
            if file.endswith(".sformat"):
                unpack_file("sformat", outpath, filepath)
            if file.endswith(".ess"):
                unpack_file("ess", outpath, filepath)

